# -*- coding: utf-8 -*-

"""Top-level package for NLPeasy."""

__author__ = """Philipp Thomann"""
__email__ = 'philipp.thomann@d-one.ai'
__version__ = '0.6.1'

from .pipeline import *
from .docker import *
from .elastic import *
from .html import *
from . import util
